// DOM-IGNORE-BEGIN
/*******************************************************************************
Copyright 2015 Microchip Technology Inc. (www.microchip.com)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

To request to license the code under the MLA license (www.microchip.com/mla_license), 
please contact mla_licensing@microchip.com
*******************************************************************************/
//DOM-IGNORE-END

/*******************************************************************************
 Change History:
 Revision     Description
 v2.6         Removed state machine 'SUBSTATE_SET_CONTROL_LINE_STATE' from normal
              enumeration process.
              Modified code to accommodate CDC devices that do not have distinct
              Communication and Data interfaces.
 v2.7         Modified the code to allow connection of USB-RS232 dongles that do
              not fully comply with CDC specifications
              Modified API USBHostCDC_Api_Send_OUT_Data to allow data transfers
              more than 256 bytes
 v2.9         Fixed bug with sending SET_CONTROL_LINE_STATE command to device.
              The bit map settings were being sent as data stage bytes, rather
			  than as the wValue of the SETUP packet.
********************************************************************************/
/*  ̃\[X́AUSBzXg֘Âbcbs^XNCut@CłB

    2019-02-05  ޒH[(ނ)  Rg̓{ꉻ(Japanese commentary)
 */

#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include "usb_config.h"
#include "usb.h"
#include "usb_host_cdc.h"
#include "usb_host.h"

//#define DEBUG_MODE   // fobO\Lɂ܂
#ifdef DEBUG_MODE
    #include "../skDEBUG.h"
#endif

#if defined(USB_NUM_BULK_NAKS)
    #define USB_HOST_CDC_NAK_TIMEOUT 1
#else
    #define USB_HOST_CDC_NAK_TIMEOUT 0
#endif

// *****************************************************************************
// *****************************************************************************
// Configuration
// *****************************************************************************
// *****************************************************************************

// *****************************************************************************
/* T|[gĂfoCX̍ő吔

̒ĺÃNXhCoT|[głőڑfoCX\܂B
[U[l`Ȃꍇ́A1ɐݒ肳܂B
USBzXgC̐ɂA݂1ɐݒ肷Kv܂B
*/
#ifndef USB_MAX_CDC_DEVICES
    #define USB_MAX_CDC_DEVICES        1
#endif

// *****************************************************************************
// *****************************************************************************
// Constants
// *****************************************************************************
// *****************************************************************************

// *****************************************************************************
// State Machine Constants
// *****************************************************************************

#ifndef USB_ENABLE_TRANSFER_EVENT

#define STATE_MASK                                0x00F0 //
#define SUBSTATE_MASK                             0x000F //
                                                   
#define NEXT_STATE                                0x0010 //
#define NEXT_SUBSTATE                             0x0001 //


#define STATE_DETACHED                            0x0000 //
                                                   
#define STATE_INITIALIZE_DEVICE                   0x0010 //
#define SUBSTATE_WAIT_FOR_ENUMERATION             0x0000 //
#define SUBSTATE_DEVICE_ENUMERATED                0x0001 //
#define SUBSTATE_GET_LINE_CODING                  0x0002 //
#define SUBSTATE_WAIT_FOR_GET_LINE_CODING         0x0003 //
#define SUBSTATE_SET_LINE_CODING                  0x0004 //
#define SUBSTATE_WAIT_FOR_SET_LINE_CODING         0x0005 //
#define SUBSTATE_SET_CONTROL_LINE_STATE           0x0006 //
#define SUBSTATE_WAIT_FOR_SET_CONTROL_LINE_STATE  0x0007 //


#define STATE_RUNNING                             0x0020 //
#define SUBSTATE_WAITING_FOR_REQ                  0x0000 //
#define SUBSTATE_SEND_READ_REQ                    0x0001 //
#define SUBSTATE_READ_REQ_WAIT                    0x0002 //
#define SUBSTATE_SEND_WRITE_REQ                   0x0003 //
#define SUBSTATE_WRITE_REQ_WAIT                   0x0004 //

#define STATE_CDC_RESET_RECOVERY                  0x0030 //
#define SUBSTATE_SEND_RESET                       0x0000 //
#define SUBSTATE_WAIT_FOR_RESET                   0x0001 //
#define SUBSTATE_RESET_COMPLETE                   0x0002 //
                                                  
#define STATE_HOLDING                             0x0040 //


#else


#define STATE_DETACHED                            0x0000 //

#define STATE_INITIALIZE_DEVICE                   0x0001 //
#define STATE_WAIT_FOR_GET_LINE_CODING            0x0002 //
#define STATE_WAIT_FOR_SET_LINE_CODING            0x0003 //
#define STATE_WAIT_FOR_SET_CONTROL_LINE_STATE     0x0004 //

#define STATE_RUNNING                             0x0005 //
#define STATE_READ_REQ_WAIT                       0x0006 //
#define STATE_WRITE_REQ_WAIT                      0x0007 //

#define STATE_WAIT_FOR_RESET                      0x0008 //
#define STATE_RESET_COMPLETE                      0x0009 //

#define STATE_HOLDING                             0x000A //

#endif

// *****************************************************************************
// Other Constants
// *****************************************************************************

#define USB_CDC_RESET           (0xFF)  // foCXZbg邽߂̃foCXvR[hB
#define MARK_RESET_RECOVERY     (0x0E)  // 


// RS232ݒf[^̃ftHgl
#ifndef USB_CDC_BAUDRATE_SUPPORTED 
    #define USB_CDC_BAUDRATE_SUPPORTED 19200UL      // foCX̃ftHg̒ʐMx
#endif

#ifndef USB_CDC_PARITY_TYPE 
    #define USB_CDC_PARITY_TYPE 0                   // ftHg̃peB͖
#endif

#ifndef USB_CDC_STOP_BITS 
    #define USB_CDC_STOP_BITS 0                     // ftHg̃Xgbvrbg
#endif

#ifndef USB_CDC_NO_OF_DATA_BITS 
    #define USB_CDC_NO_OF_DATA_BITS 8               // ftHg̃f[^͂Wrbg
#endif

//******************************************************************************
//******************************************************************************
// Data Structures
//******************************************************************************
//******************************************************************************

//******************************************************************************
//******************************************************************************
// Section: Local Prototypes
//******************************************************************************
//******************************************************************************
void _USBHostCDC_ResetStateJump( uint8_t i );
void USBHostCDC_Init_CDC_Buffers(void);

//******************************************************************************
//******************************************************************************
// Macros
//******************************************************************************
//******************************************************************************

#ifndef USB_ENABLE_TRANSFER_EVENT

#define _USBHostCDC_SetNextState()              { deviceInfoCDC[i].state = (deviceInfoCDC[i].state & STATE_MASK) + NEXT_STATE; }
#define _USBHostCDC_SetNextSubState()           { deviceInfoCDC[i].state += NEXT_SUBSTATE; }
#define _USBHostCDC_TerminateTransfer( error )  {                                                                       \
                                                    deviceInfoCDC[i].errorCode = error;                                 \
                                                    deviceInfoCDC[i].state     = STATE_RUNNING;\
                                                }

#else
    #define _USBHostCDC_TerminateTransfer( error )  {                                                                       \
                                                        deviceInfoCDC[i].errorCode = error;                                 \
                                                        deviceInfoCDC[i].state     = STATE_RUNNING;\
                                                    }
#endif


//******************************************************************************
//******************************************************************************
// Section: CDC Host Global Variables
//******************************************************************************
//******************************************************************************

USB_CDC_LINE_CODING           CDC_DEV_LINE_CODING_Buffer; // foCXMLINE CODINGf[^i[܂
USB_CDC_LINE_CODING           CDC_LINE_CODING_Buffer;     // zXgŐݒ肳ꂽLINE CODINGf[^i[܂
USB_CDC_CONTROL_SIGNAL_BITMAP CDC_CONTROL_SIGNAL_Buffer;  // zXgɐݒ肳ꂽCONTROL SIGNALf[^i[܂B
// ̍\̂̓Gj[VɎWꂽfoCXׂ̂Ă̏ێ܂
USB_CDC_DEVICE_INFO           deviceInfoCDC[USB_MAX_CDC_DEVICES] __attribute__ ((aligned));

uint8_t CDCdeviceAddress = 0; // ڑfoCX̃AhXێ

//******************************************************************************
//******************************************************************************
// Section: CDC Host External Variables
//******************************************************************************
//******************************************************************************

#define PTR_HOST_LINE_CODING_BUFFER  (&CDC_LINE_CODING_Buffer)    // CR[fBOobt@̃AhX
#define PTR_HOST_CONTROL_LINE_BUFFER (&CDC_CONTROL_SIGNAL_Buffer) // Mobt@̃AhX


extern CLIENT_DRIVER_TABLE usbDeviceInterfaceTable;

//******************************************************************************
//******************************************************************************
// Section: CDC Host Global Variables
//******************************************************************************
//******************************************************************************

// *****************************************************************************
// *****************************************************************************
// CDC Host Stack Callback Functions
// *****************************************************************************
// *****************************************************************************

/*******************************************************************************
  Function:
    bool USBHostCDCInitAddress( uint8_t address, uint32_t flags, uint8_t clientDriverID )

  O:
    foCX̓G[ȂŃGj[VĂ鎖B

  Tv:
    ̋@\͐ڑꂽCDCfoCX̃AhX܂B

  :
    foCXG[ȂŃGj[VƁACDCNCAg͂̊֐Ăяo܂B
    SĂ̓]vɑ΂āÃAhXCDCfoCXʂׂɎgp܂B


  :
    uint8_t address        -   VfoCX̃AhX
    uint32_t flags         -   tO
    uint8_t clientDriverID -   foCXvɑ΂NCAghCo̎ID

  ߂l:
    true    -   foCXT|[g鎖ł܂B
    false   -   T|[gł܂B

  Remarks:
    None
*******************************************************************************/
bool USBHostCDCInitAddress( uint8_t address, uint32_t flags, uint8_t clientDriverID )
{
    #ifndef MINIMUM_BUILD
//        UART2PrintString( "CDC: Device attached.\r\n" );
    #endif

    if (CDCdeviceAddress == 0)
    {
        // VfoCX̃AhXۑ܂B
        CDCdeviceAddress = address;
        return true;
    }
    else
    {
        // ݈̂1̃foCXłB
        return false;
    }        
}    


// *****************************************************************************
// *****************************************************************************
// Application Callable Functions
// *****************************************************************************
// *****************************************************************************

/*******************************************************************************
  Function:
    uint8_t    USBHostCDCDeviceStatus( uint8_t deviceAddress )

  Summary:

  :
    ̋@\CDCfoCX̃Xe[^X(Ԃ𓾂)܂B

  Preconditions:  None

  :
    uint8_t deviceAddress - ₢킹foCX̃AhX

  ߂l:
    USB_CDC_DEVICE_NOT_FOUND         -  foCXAhXȂAfoCXCDCł͂܂
    USB_CDC_INITIALIZING             -  CDĆAڑƏ̉ߒɂ܂
    USB_PROCESSING_REPORT_DESCRIPTOR -  CDCfoCXoA|[gfBXNv^͂Ă܂
    USB_CDC_NORMAL_RUNNING           -  CDCfoCX͐ɓ삵ĂA|[g𑗎M鏀łĂ܂
    USB_CDC_DEVICE_HOLDING           -  CDCfoCX̓G[񕜂ł܂ł
    USB_CDC_DEVICE_DETACHED          -  CDC؂藣܂

  Remarks:
    None
*******************************************************************************/
uint8_t    USBHostCDCDeviceStatus( uint8_t deviceAddress )
{
    uint8_t    i;
    uint8_t    status;

    // w̃foCX܂B
    for (i=0; (i<USB_MAX_CDC_DEVICES) && (deviceInfoCDC[i].deviceAddress != deviceAddress); i++);
    if (i == USB_MAX_CDC_DEVICES)
    {
        return USB_CDC_DEVICE_NOT_FOUND;     // w肳ꂽAhX̋@͗pł܂B
    }
    // foCX݂̌̃Xe[^X𓾂
    status = USBHostDeviceStatus( deviceAddress );
    if (status != USB_DEVICE_ATTACHED)
    {
        return status;   // foCX͐ڑĂȂғłȂꍇ͂Ŗ߂
    }
    else
    {
        // foCXtăGj[V܂B  
        // Ƌ̓IȃXe[^X擾鎖ł܂B
        #ifndef USB_ENABLE_TRANSFER_EVENT
           switch (deviceInfoCDC[i].state & STATE_MASK)
           {
               case STATE_INITIALIZE_DEVICE:
                   return USB_CDC_INITIALIZING;
                   break;

               case STATE_RUNNING:
                   return USB_CDC_NORMAL_RUNNING;
                   break;
               case STATE_HOLDING:
                   return USB_CDC_DEVICE_HOLDING;
                   break;

               case STATE_CDC_RESET_RECOVERY:
                   return USB_CDC_RESETTING_DEVICE;
                   break;
                   
               default:  
                   return USB_CDC_DEVICE_DETACHED;  
                   break;
           }
        #else
           switch (deviceInfoCDC[i].state)
           {
               case STATE_INITIALIZE_DEVICE:
                   return USB_CDC_INITIALIZING;   // foCX͏łB
                   break;

               case STATE_RUNNING:
               case STATE_READ_REQ_WAIT:
               case STATE_WRITE_REQ_WAIT:
                   return USB_CDC_NORMAL_RUNNING; // foCX͎sŁAf[^]Ɏgpł܂B
                   break;

               case STATE_HOLDING:
                   return USB_CDC_DEVICE_HOLDING; // G[ׂ̈ɃfoCX͕ۗłB
                   break;
                   
               case STATE_WAIT_FOR_RESET:
               case STATE_RESET_COMPLETE:
                   return USB_CDC_RESETTING_DEVICE; // foCX̓ZbgłB
                   break;
               default:  // Other various STATE
                   return USB_CDC_DEVICE_DETACHED;  // foCX͐؂藣Ă܂B
                   break;
           }
        #endif
    }    
}
    
/*******************************************************************************
  Function:
    uint8_t USBHostCDCResetDevice( uint8_t deviceAddress )

  Tv:
    ̋@\CDCZbgJn܂B

  :
    ZbǵAuڑĂďĂȂꍇɂ̂ݔsł܂B

  Precondition:
    None

  :
    uint8_t deviceAddress - ZbgfoCXAhX

  ߂l:
    USB_SUCCESS                 - ZbgJn
    USB_MSD_DEVICE_NOT_FOUND    - w肳ꂽAhXfoCX܂
    USB_MSD_ILLEGAL_REQUEST     - foCX̓Zbgɑ΂ĕsȏԂł

  Remarks:
    None
*******************************************************************************/
uint8_t USBHostCDCResetDevice( uint8_t deviceAddress )
{
    uint8_t    i;

    // LȃfoCXvĂ鎖mF܂B
    if ((deviceAddress == 0) || (deviceAddress > 127))
    {
        return USB_CDC_DEVICE_NOT_FOUND;
    }

    // foCX܂B
    for (i=0; (i<USB_MAX_CDC_DEVICES) && (deviceInfoCDC[i].deviceAddress != deviceAddress); i++);
    if (i == USB_MAX_CDC_DEVICES)
    {
        return USB_CDC_DEVICE_NOT_FOUND;
    }

     #ifndef USB_ENABLE_TRANSFER_EVENT
       if (((deviceInfoCDC[i].state & STATE_MASK) != STATE_DETACHED) &&
            ((deviceInfoCDC[i].state & STATE_MASK) != STATE_INITIALIZE_DEVICE))
     #else
         if ((deviceInfoCDC[i].state != STATE_DETACHED) &&
             (deviceInfoCDC[i].state != STATE_INITIALIZE_DEVICE))
     #endif
        {
            deviceInfoCDC[i].flags.val |= MARK_RESET_RECOVERY;
            deviceInfoCDC[i].flags.bfReset = 1;
        #ifndef USB_ENABLE_TRANSFER_EVENT
            deviceInfoCDC[i].returnState = STATE_HOLDING;
        #else
            deviceInfoCDC[i].returnState = STATE_RUNNING;
        #endif
        
            _USBHostCDC_ResetStateJump( i );
            return USB_SUCCESS;
        }
        return USB_CDC_ILLEGAL_REQUEST;
}


/*******************************************************************************
  Function:
     void USBHostCDCTasks( void )

  Tv:
    ̋@\́ACDCNXɕKvȃeiX^XNs܂B

  :
    zXgw̓]CxggpĂȂꍇ́A
    AvP[VɂĒIɌĂяoKv܂B
    zXgw̓]CxggpĂꍇA
    ̊֐̓RpCĂ̂ŁAĂяoKv͂܂B

  O:
    USBHostCDCInitialize() ĂяoĒu܂B

  Parameters:
    None - None

  Returns:
    None

  Remarks:
    None
*******************************************************************************/
void USBHostCDCTasks( void )
{
#ifndef USB_ENABLE_TRANSFER_EVENT
    uint32_t   byteCount;
    uint8_t    errorCode;
    uint8_t    i;

//    uint8_t    temp;

    for (i=0; i<USB_MAX_CDC_DEVICES; i++)
    {
        if (deviceInfoCDC[i].deviceAddress == 0) /* ʑwɂčXVꂽfoCXAhX */
        {
            deviceInfoCDC[i].state = STATE_DETACHED;
        }

        switch (deviceInfoCDC[i].state & STATE_MASK)
        {
            case STATE_DETACHED:
                // foCXڑĂ܂B
                break;

            case STATE_INITIALIZE_DEVICE:
                switch (deviceInfoCDC[i].state & SUBSTATE_MASK)
                {
                    case SUBSTATE_WAIT_FOR_ENUMERATION:
                        if (USBHostDeviceStatus( deviceInfoCDC[i].deviceAddress ) == USB_DEVICE_ATTACHED)
                        {
                            _USBHostCDC_SetNextSubState();
                        }
                        break;

                    case SUBSTATE_DEVICE_ENUMERATED:
                        // AvP[VpɃfoCXAhX
                        USBHostCDCInitAddress( deviceInfoCDC[i].deviceAddress,0, deviceInfoCDC[i].clientDriverID );
                        // Set ConfigSubԂǉALANGIDString fBXNv^擾Kv܂
                        _USBHostCDC_SetNextSubState();
                        break;


                   case SUBSTATE_GET_LINE_CODING:
                         
                       if(!USBHostIssueDeviceRequest( deviceInfoCDC[i].deviceAddress, USB_SETUP_DEVICE_TO_HOST | USB_SETUP_TYPE_CLASS | USB_SETUP_RECIPIENT_INTERFACE, 
                                            USB_CDC_GET_LINE_CODING, 0 , deviceInfoCDC[i].commInterface.interfaceNum , USB_CDC_LINE_CODING_LENGTH, (uint8_t*)&CDC_DEV_LINE_CODING_Buffer,
                                            USB_DEVICE_REQUEST_GET , deviceInfoCDC[i].clientDriverID ))
                       {           
                            _USBHostCDC_SetNextSubState();
                       }
                        break;

                    case SUBSTATE_WAIT_FOR_GET_LINE_CODING:   
                        if (USBHostTransferIsComplete( deviceInfoCDC[i].deviceAddress, 0, &errorCode, &byteCount ))
                        {
                            if (errorCode)
                            {
                                /* G[R[hݒ肷 */
                                _USBHostCDC_TerminateTransfer(errorCode);
                            }
                            else
                            {
                                // STALLNA܂AEP0ł̂ŃXg[NAKv͂܂B
                                USBHostClearEndpointErrors( deviceInfoCDC[i].deviceAddress, 0 );
                                // zXgŗ\zR[fBOƃfoCXM̂r܂B
                                if(0 == memcmp(&CDC_DEV_LINE_CODING_Buffer, PTR_HOST_LINE_CODING_BUFFER, USB_CDC_LINE_CODING_LENGTH))
                                {
                                    // 䃉C̏Ԃׂݒ肵ꍇ
 //                                   deviceInfoCDC[i].state = STATE_INITIALIZE_DEVICE | SUBSTATE_SET_CONTROL_LINE_STATE;
                                      deviceInfoCDC[i].state = STATE_RUNNING;
                                }
                                else
                                {
                                    // f[^\zǂłȂꍇ́AR[fBO̐ݒɐi݂܂B
                                    deviceInfoCDC[i].state = STATE_INITIALIZE_DEVICE | SUBSTATE_SET_LINE_CODING;
                                }
                            }
                        }
                        break;

                   case SUBSTATE_SET_LINE_CODING:
                       // R[fBO̐ݒ
                       if(!USBHostIssueDeviceRequest( deviceInfoCDC[i].deviceAddress, USB_SETUP_HOST_TO_DEVICE | USB_SETUP_TYPE_CLASS | USB_SETUP_RECIPIENT_INTERFACE, 
                                            USB_CDC_SET_LINE_CODING, 1 , deviceInfoCDC[i].commInterface.interfaceNum , USB_CDC_LINE_CODING_LENGTH, (uint8_t*)PTR_HOST_LINE_CODING_BUFFER,
                                            USB_DEVICE_REQUEST_SET , deviceInfoCDC[i].clientDriverID ))
                       {           
                            _USBHostCDC_SetNextSubState();
                       }
                        break;

                    case SUBSTATE_WAIT_FOR_SET_LINE_CODING:   
                        if (USBHostTransferIsComplete( deviceInfoCDC[i].deviceAddress, 0, &errorCode, &byteCount ))
                        {
                            if (errorCode)
                            {
                                /* G[R[hݒ肷 */
                                _USBHostCDC_TerminateTransfer(errorCode);
                            }
                            else
                            {    // R[fBOݒ肳Ă邩ēxmF܂B
                                 deviceInfoCDC[i].state = STATE_INITIALIZE_DEVICE | SUBSTATE_GET_LINE_CODING;
                            }
                        }
                        break;

                   case SUBSTATE_SET_CONTROL_LINE_STATE:
                         
                       if(!USBHostIssueDeviceRequest( deviceInfoCDC[i].deviceAddress, USB_SETUP_HOST_TO_DEVICE | USB_SETUP_TYPE_CLASS | USB_SETUP_RECIPIENT_INTERFACE, 
                                            USB_CDC_SET_CONTROL_LINE_STATE , *(uint16_t*)PTR_HOST_CONTROL_LINE_BUFFER , deviceInfoCDC[i].commInterface.interfaceNum , USB_CDC_CONTROL_LINE_LENGTH, NULL,
                                            USB_DEVICE_REQUEST_SET , deviceInfoCDC[i].clientDriverID ))
                       {           
                            _USBHostCDC_SetNextSubState();
                       }
                        break;

                    case SUBSTATE_WAIT_FOR_SET_CONTROL_LINE_STATE:   
                        if (USBHostTransferIsComplete( deviceInfoCDC[i].deviceAddress, 0, &errorCode, &byteCount ))
                        {
                            if (errorCode)
                            {
                                /* G[R[hݒ肷 */
                                _USBHostCDC_TerminateTransfer(errorCode);
                            }
                            else
                            {
                                _USBHostCDC_SetNextState();
                            }
                        }
                        break;

                    default : 
                        break;
                }
                break;

            case STATE_RUNNING:
                switch (deviceInfoCDC[i].state & SUBSTATE_MASK)
                {
                    case SUBSTATE_WAITING_FOR_REQ:   
                              /* AvP[V̗v҂Ă܂ */
                        break;

                    case SUBSTATE_SEND_READ_REQ:   
                            // Gh|Cg0Rg[]gpꍇA
                            // ȊO̒ʐMC^tF[Xv̓f[^C^tF[X]płB
                            USBHostClearEndpointErrors( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA );

                            if(deviceInfoCDC[i].endpointDATA == 0x00)
                            {
                                errorCode   = USBHostIssueDeviceRequest( deviceInfoCDC[i].deviceAddress, USB_SETUP_DEVICE_TO_HOST | USB_SETUP_TYPE_CLASS | USB_SETUP_RECIPIENT_INTERFACE, 
                                                                    deviceInfoCDC[i].commRequest, 0, deviceInfoCDC[i].interface,deviceInfoCDC[i].reportSize, deviceInfoCDC[i].userData,
                                                                    USB_DEVICE_REQUEST_GET , deviceInfoCDC[i].clientDriverID );
                            }
                            else
                            {
                                errorCode = USBHostRead( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA,
                                                  deviceInfoCDC[i].userData, deviceInfoCDC[i].reportSize );
                            }

                            if (errorCode)
                            {
                                if(USB_ENDPOINT_STALLED == errorCode)
                                {
                                     USBHostClearEndpointErrors( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA );
                                     deviceInfoCDC[i].returnState = STATE_HOLDING;
                                     deviceInfoCDC[i].flags.bfReset = 1;
                                     _USBHostCDC_ResetStateJump( i );
                                }
                                else
                                {
                                    _USBHostCDC_TerminateTransfer(errorCode);
                                }
                            }
                            else
                            {
                                _USBHostCDC_SetNextSubState();
                            }
                        break;

                    case SUBSTATE_READ_REQ_WAIT:   
                              if (USBHostTransferIsComplete( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA, &errorCode, &byteCount ))
                              {
                                if (errorCode)
                                {
                                    if(USB_ENDPOINT_STALLED == errorCode)
                                    {
                                         USBHostClearEndpointErrors( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA );
                                         deviceInfoCDC[i].returnState = STATE_HOLDING;
                                         deviceInfoCDC[i].flags.bfReset = 1;
                                         _USBHostCDC_ResetStateJump( i );
                                    }else if(USB_ENDPOINT_NAK_TIMEOUT == errorCode)
                                    {
                                        USB_HOST_APP_EVENT_HANDLER(deviceInfoCDC[i].deviceAddress,EVENT_CDC_NAK_TIMEOUT,NULL, 0);
                                        deviceInfoCDC[i].state = STATE_RUNNING | SUBSTATE_WAITING_FOR_REQ;
                                        USBHostClearEndpointErrors( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA );
                                    }
                                    else
                                    {
                                        _USBHostCDC_TerminateTransfer(errorCode);
                                    }
                                }
                                else
                                {
                                    // STALLNA܂AEP0ł̂ŃXg[NAKv͂܂B
                                    USBHostClearEndpointErrors( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA );
                                    deviceInfoCDC[i].bytesTransferred = byteCount; /* |[g̃TCYƃtOG[Ŕr鎖ł܂H*/
                                    deviceInfoCDC[i].state = STATE_RUNNING | SUBSTATE_WAITING_FOR_REQ;
                                }
                              }
#ifdef DEBUG_MODE
                              UART2PrintString("|");
#endif                                
                        break;

                    case SUBSTATE_SEND_WRITE_REQ:   
                              // Gh|Cg0Rg[]gpꍇA
                              // ȊO̒ʐMC^tF[Xv̓f[^C^tF[X]płB
                              if(deviceInfoCDC[i].endpointDATA == 0x00)
                              {
                                  errorCode   = USBHostIssueDeviceRequest( deviceInfoCDC[i].deviceAddress, USB_SETUP_HOST_TO_DEVICE | USB_SETUP_TYPE_CLASS | USB_SETUP_RECIPIENT_INTERFACE, 
                                               deviceInfoCDC[i].commRequest, 0, deviceInfoCDC[i].interface,deviceInfoCDC[i].reportSize, deviceInfoCDC[i].userData,
                                               USB_DEVICE_REQUEST_SET , deviceInfoCDC[i].clientDriverID );
                              }
                              else
                              {   // ]TCY64oCg𒴂ꍇ
                                  // f[^]ɕ̃gUNVgpĂ
                                  if(deviceInfoCDC[i].remainingBytes > USB_CDC_MAX_PACKET_SIZE)
                                  {
                                    deviceInfoCDC[i].bytesTransferred = USB_CDC_MAX_PACKET_SIZE; // őgUNVTCY64oCgł
                                    deviceInfoCDC[i].remainingBytes -= USB_CDC_MAX_PACKET_SIZE;
                                    errorCode = USBHostWrite( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA,
                                                    deviceInfoCDC[i].userData, deviceInfoCDC[i].bytesTransferred );
                                    deviceInfoCDC[i].userData = deviceInfoCDC[i].userData + USB_CDC_MAX_PACKET_SIZE;
                                  }
                                  else
                                  {
                                    deviceInfoCDC[i].bytesTransferred = deviceInfoCDC[i].remainingBytes;
                                    deviceInfoCDC[i].remainingBytes = 0;
                                    errorCode = USBHostWrite( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA,
                                                    deviceInfoCDC[i].userData, deviceInfoCDC[i].bytesTransferred );
                                  }

                              }

                              if (errorCode)
                              {
                                  if(USB_ENDPOINT_STALLED == errorCode)
                                  {
                                       USBHostClearEndpointErrors( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA );
                                       deviceInfoCDC[i].returnState = STATE_HOLDING;
                                       deviceInfoCDC[i].flags.bfReset = 1;
                                       _USBHostCDC_ResetStateJump( i );
                                  }
                                  else
                                  {
                                      _USBHostCDC_TerminateTransfer(errorCode);
                                  }
                              }
                              else
                              {
                                  _USBHostCDC_SetNextSubState();
                              }

                        break;
                    case SUBSTATE_WRITE_REQ_WAIT:   
                              if (USBHostTransferIsComplete( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA, &errorCode, &byteCount ))
                              {
                                if (errorCode)
                                {
                                    if(USB_ENDPOINT_STALLED == errorCode)
                                    {
                                         USBHostClearEndpointErrors( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA );
                                         deviceInfoCDC[i].returnState = STATE_HOLDING;
                                         deviceInfoCDC[i].flags.bfReset = 1;
                                         _USBHostCDC_ResetStateJump( i );
                                    }
                                    else
                                    {
                                        _USBHostCDC_TerminateTransfer(errorCode);
                                    }
                                }
                                else
                                {
                                    USBHostClearEndpointErrors( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA );
                                    // ͓]ׂc̃f[^oCg邩ǂmF܂
                                    if(deviceInfoCDC[i].remainingBytes != 0)
                                    {
                                        // ʂOUTvZbgAbv
                                        deviceInfoCDC[i].state = STATE_RUNNING | SUBSTATE_SEND_WRITE_REQ;
                                    }
                                    else
                                    {
                                        // łȂΑ҂Ԃ
                                        deviceInfoCDC[i].state = STATE_RUNNING | SUBSTATE_WAITING_FOR_REQ;
                                    }
                                }
                              }

                        break;

                    default : 
                        break;
                }
                break;

            case STATE_CDC_RESET_RECOVERY:
                switch (deviceInfoCDC[i].state & SUBSTATE_MASK)
                {
                    case SUBSTATE_SEND_RESET:   /* Zbgv킩Ȃ */
                            errorCode = USBHostIssueDeviceRequest( deviceInfoCDC[i].deviceAddress, USB_SETUP_HOST_TO_DEVICE | USB_SETUP_TYPE_CLASS | USB_SETUP_RECIPIENT_INTERFACE,
                                            USB_CDC_RESET, 0, deviceInfoCDC[i].interface, 0, NULL, USB_DEVICE_REQUEST_SET , deviceInfoCDC[i].clientDriverID );

                            if (errorCode)
                            {
                                    if(USB_ENDPOINT_STALLED == errorCode)
                                    {
                                         USBHostClearEndpointErrors( deviceInfoCDC[i].deviceAddress, 0 );
                                         deviceInfoCDC[i].returnState = STATE_HOLDING;
                                         deviceInfoCDC[i].flags.bfReset = 1;
                                         _USBHostCDC_ResetStateJump( i );
                                    }
                                    else
                                    {
                                        _USBHostCDC_TerminateTransfer(errorCode);
                                    }
                            }
                            else
                            {
                                _USBHostCDC_SetNextSubState();
                            }
                        break;
                    case SUBSTATE_WAIT_FOR_RESET:
                            if (USBHostTransferIsComplete( deviceInfoCDC[i].deviceAddress, 0, &errorCode, &byteCount ))
                            {
                                if (errorCode)
                                {
                                    if(USB_ENDPOINT_STALLED == errorCode)
                                    {
                                         USBHostClearEndpointErrors( deviceInfoCDC[i].deviceAddress, 0 );
                                         deviceInfoCDC[i].returnState = STATE_HOLDING;
                                         deviceInfoCDC[i].flags.bfReset = 1;
                                         _USBHostCDC_ResetStateJump( i );
                                    }
                                    else
                                    {
                                        _USBHostCDC_TerminateTransfer(errorCode);
                                    }
                                }
                                else
                                {
                                    deviceInfoCDC[i].flags.bfReset = 0;
                                    _USBHostCDC_ResetStateJump( i );
                                }
                            }
                            break;

                    case SUBSTATE_RESET_COMPLETE:
                            _USBHostCDC_ResetStateJump( i );
                            break;
                }
                break;

            case STATE_HOLDING:
                break;

            default : 
                break;

        }
    }
#endif
}


/*******************************************************************************
  Function:
    USBHostCDCTransfer( uint8_t deviceAddress, uint8_t direction, uint8_t reportid,
                        uint8_t size, uint8_t *data)

  Tv:
    ̋@\CDC]Jn܂B

  :
    ̋@\CDC]Jn܂B
    ̋@\ɃANZXׂɁAǂݎ/݃bp[AvP[VC^[tFCXt@CɗpӂĂ܂B
    ̋@\́AʐMC^tF[Xƃf[^C^tF[X̗]vɎgp܂B

  Preconditions:
    None

  :
    uint8_t deviceAddress      - foCXAhX
    uint8_t direction          - 1=ǂݎ, 0=
    uint8_t interfaceNum       - foCX̃C^[tFCXԍ
    uint8_t size               - f[^obt@̃oCgTCY
    uint8_t *data              - f[^obt@ւ̃|C^
    uint8_t endpointDATA       - ]vꂽGh|Cg̏ڍ

  ߂l:
    USB_SUCCESS                 - NGXg͐ɊJn܂
    USB_CDC_DEVICE_NOT_FOUND    - w肳ꂽAhXfoCX܂
    USB_CDC_DEVICE_BUSY         - ]sׂ̃foCXK؂ȏԂł͂܂

  Remarks:
    None
*******************************************************************************/

uint8_t USBHostCDCTransfer( uint8_t deviceAddress,uint8_t request , uint8_t direction, uint8_t interfaceNum, uint16_t size, uint8_t *data , uint8_t endpointDATA)
{
    uint8_t    i;
    #ifdef USB_ENABLE_TRANSFER_EVENT
    uint8_t    errorCode;
    #endif
    // foCXĂB
    for (i=0; (i<USB_MAX_CDC_DEVICES) && (deviceInfoCDC[i].deviceAddress != deviceAddress); i++);
    if (i == USB_MAX_CDC_DEVICES)
    {
        return USB_CDC_DEVICE_NOT_FOUND;
    }

    // foCXǂݏ\ȏԂɂȂĂ邱ƂmFĂB
    #ifndef USB_ENABLE_TRANSFER_EVENT
        if (deviceInfoCDC[i].state != (STATE_RUNNING) &&
           (deviceInfoCDC[i].state & SUBSTATE_MASK) != (SUBSTATE_WAITING_FOR_REQ))
    #else
        if (deviceInfoCDC[i].state != STATE_RUNNING)
    #endif
        {
            return USB_CDC_DEVICE_BUSY;
        }
     
    // ]܂B
    deviceInfoCDC[i].bytesTransferred  = 0;
    deviceInfoCDC[i].errorCode         = USB_SUCCESS;
    deviceInfoCDC[i].userData          = data;
    deviceInfoCDC[i].reportSize        = size;
    deviceInfoCDC[i].remainingBytes    = size;
    deviceInfoCDC[i].interface         = interfaceNum;
    deviceInfoCDC[i].endpointDATA      = endpointDATA;
    deviceInfoCDC[i].commRequest       = request;  // DATA]vꂽꍇ͖ȃGg
    #ifdef DEBUG_MODE
        UART2PrintString( "Data EP: " );
        UART2PutHex( deviceInfoCDC[i].endpointDATA );
        UART2PrintString( "\r\n" );
    #endif
    
    #ifndef USB_ENABLE_TRANSFER_EVENT
        // ]ԂɃWv܂B
        if(!direction)
            {
                /* ݗv𑗐M */
                deviceInfoCDC[i].state = STATE_RUNNING | SUBSTATE_SEND_WRITE_REQ;
            }
        else
            {
                /* ǂݎv𑗐M */
                deviceInfoCDC[i].state = STATE_RUNNING | SUBSTATE_SEND_READ_REQ;
            }
    #else
        // ̊֐̓Gh|Cg̓G[Ԃ܂B
        USBHostClearEndpointErrors( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA );

        if(!direction)
            {
                if(deviceInfoCDC[i].endpointDATA == 0x00) // Gh|Cg0COMMC^[tF[Xvɐ]gpꍇ
                {
                    errorCode = USBHostIssueDeviceRequest( deviceInfoCDC[i].deviceAddress, USB_SETUP_HOST_TO_DEVICE | USB_SETUP_TYPE_CLASS | USB_SETUP_RECIPIENT_INTERFACE, 
                                                        request, 0, deviceInfoCDC[i].interface,deviceInfoCDC[i].reportSize, deviceInfoCDC[i].userData,
                                                        USB_DEVICE_REQUEST_SET , deviceInfoCDC[i].clientDriverID );
                    deviceInfoCDC[i].state = STATE_WRITE_REQ_WAIT;
                }
                else
                {
                    errorCode = USBHostWrite( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA,
                                              deviceInfoCDC[i].userData, deviceInfoCDC[i].reportSize );
                    deviceInfoCDC[i].state = STATE_WRITE_REQ_WAIT;
                }
            }
        else
            {
                if(deviceInfoCDC[i].endpointDATA == 0x00) // Gh|Cg0COMMC^[tF[Xvɐ]gpꍇ
                {
                    errorCode = USBHostIssueDeviceRequest( deviceInfoCDC[i].deviceAddress, USB_SETUP_DEVICE_TO_HOST | USB_SETUP_TYPE_CLASS | USB_SETUP_RECIPIENT_INTERFACE, 
                                                        request, 0, deviceInfoCDC[i].interface,deviceInfoCDC[i].reportSize, deviceInfoCDC[i].userData,
                                                        USB_DEVICE_REQUEST_GET , deviceInfoCDC[i].clientDriverID );
                    deviceInfoCDC[i].state = STATE_READ_REQ_WAIT;
                }
                else
                {
                    errorCode = USBHostRead( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA, 
                                             deviceInfoCDC[i].userData, deviceInfoCDC[i].reportSize );
                    deviceInfoCDC[i].state = STATE_READ_REQ_WAIT;
                }
            }

        if(errorCode)
            {
                _USBHostCDC_TerminateTransfer( USB_CDC_RESET_ERROR );
            }
        else
            {
                deviceInfoCDC[i].flags.bfReset = 0;
            }
    #endif
    return USB_SUCCESS;
}
/*******************************************************************************
  Function:
    bool USBHostCDCTransferIsComplete( uint8_t deviceAddress,
                        uint8_t *errorCode, uint32_t *byteCount )

  Tv:
    ̋@\́AŌ̓]ǂ܂B

  :
    ̋@\́AŌ̓]ǂ܂B
    ֐trueԂꍇAԂoCgƃG[R[h͗LłB
   xɎsł]1łAgpłGh|Cg1Ȃ̂ŁA
   foCXAhXm邾ōς݂܂B

  Precondition:
    None

  :
    uint8_t deviceAddress  - foCXAhX
    uint8_t *errorCode     - Ō̓]̃G[R[h
    uint32_t *byteCount    - ]oCg

  ߂l:
    true    - ]AerrorCode͗Lł
    false   - ]͊Ă܂AG[R[h͖ł
*******************************************************************************/

bool USBHostCDCTransferIsComplete( uint8_t deviceAddress, uint8_t *errorCode, uint8_t *byteCount )
{
    uint8_t    i;

    // foCXĂB
    for (i=0; (i<USB_MAX_CDC_DEVICES) && (deviceInfoCDC[i].deviceAddress != deviceAddress); i++);
    if ((i == USB_MAX_CDC_DEVICES) || (deviceInfoCDC[i].state == STATE_DETACHED))
    {
        *errorCode = USB_CDC_DEVICE_NOT_FOUND;
        *byteCount = 0;
        return true;
    }

    *byteCount = deviceInfoCDC[i].bytesTransferred;
    *errorCode = deviceInfoCDC[i].errorCode;

      #ifndef USB_ENABLE_TRANSFER_EVENT
            if(deviceInfoCDC[i].state == (STATE_RUNNING | SUBSTATE_WAITING_FOR_REQ))
      #else
            if(deviceInfoCDC[i].state == STATE_RUNNING)
      #endif
                {
                    return true;
                }
            else
                {
                    return false;
                }
}

// *****************************************************************************
// *****************************************************************************
// Host Stack Interface Functions
// *****************************************************************************
// *****************************************************************************

/*******************************************************************************
  Function:
    bool USBHostCDCEventHandler( uint8_t address, USB_EVENT event,
                        void *data, uint32_t size )

  O:
    foCXĂ鎖

  Tv:
    ̊֐͂̃NCAghCõCxgnhłB

  :
    ܂܂ȃCxgɃzXgwɂČĂяo܂B

  :
    uint8_t address - foCX̃AhX
    USB_EVENT event - Cxg
    void *data      - CxgɊ֘Af[^ւ̃|C^
    uint32_t size   - f[^TCY

  ߂l:
    true   - Cxg܂
    false  - Cxg͏܂ł

  Remarks:
    None
*******************************************************************************/
bool USBHostCDCEventHandler( uint8_t address, USB_EVENT event, void *data, uint32_t size )
{
    HOST_TRANSFER_DATA* transfer_data = data ;
    uint8_t    i;
    #ifdef USB_ENABLE_TRANSFER_EVENT

    uint8_t    errorCode = 0 ;
    uint32_t   byteCount = 0;

    #endif

    switch (event)
    {
        case EVENT_NONE:             // Cxg͔ĂȂiNULLCxgj
            USBTasks();
            return true;
            break;

        case EVENT_DETACH:           // USBP[uOꂽidata uint8_t ŃfoCX̃AhXj
            #ifdef DEBUG_MODE
                UART2PrintString( "CDC: Detach\r\n" );
            #endif
            // e[uŃfoCXꍇ́AdvȃtB[hNAĂB
            for (i=0; (i<USB_MAX_CDC_DEVICES) && (deviceInfoCDC[i].deviceAddress != address); i++);
            if (i < USB_MAX_CDC_DEVICES)
            {
                /* CDCfoCXɂĎgpĂ郁 */
                deviceInfoCDC[i].deviceAddress = 0;
                deviceInfoCDC[i].state         = STATE_DETACHED;
                CDCdeviceAddress = 0;
                USB_HOST_APP_EVENT_HANDLER(deviceInfoCDC[i].deviceAddress,EVENT_DETACH,NULL, 0);
           }

            return true;
            break;

        case EVENT_HUB_ATTACH:       // USBnuڑꂽ
        case EVENT_TRANSFER:         // USB]܂ - gp
            #if defined( USB_ENABLE_TRANSFER_EVENT )
                #ifdef DEBUG_MODE
                    UART2PrintString( "CDC: transfer event: " );
                    UART2PutHex( address );
                    UART2PrintString( "\r\n" );
                #endif

                for (i=0; (i<USB_MAX_CDC_DEVICES) && (deviceInfoCDC[i].deviceAddress != address); i++) {}
                if (i == USB_MAX_CDC_DEVICES)
                {
                    #ifdef DEBUG_MODE
                        UART2PrintString( "CDC: Unknown device\r\n" );
                    #endif
                    return false;  // sȃfoCX
                }
                #ifdef DEBUG_MODE
                    UART2PrintString( "CDC: Device state: " );
                    UART2PutHex( deviceInfoCDC[i].state );
                    UART2PrintString( "\r\n" );
                #endif
                switch (deviceInfoCDC[i].state)
                {

                    case STATE_WAIT_FOR_GET_LINE_CODING :
                            if (USBHostTransferIsComplete( deviceInfoCDC[i].deviceAddress, 0, &errorCode, &byteCount ))
                            {
                                if (errorCode)
                                {
                                    /* G[R[hݒ肷 */
                                    _USBHostCDC_TerminateTransfer(errorCode);
                                }
                                else
                                {
                                    // STALLNA܂B
                                    // EP0ł̂ŁAX̓Xg[NAKv͂܂B
                                    USBHostClearEndpointErrors( deviceInfoCDC[i].deviceAddress, 0 );
                                    // zXgŗ\zR[fBOƃfoCXM̂r܂B
                                    if(0 == memcmp(&CDC_DEV_LINE_CODING_Buffer, PTR_HOST_LINE_CODING_BUFFER, USB_CDC_LINE_CODING_LENGTH))
                                    {
                                       // 䃉C̏Ԃׂݒ肵ꍇ
                                       if(!USBHostIssueDeviceRequest( deviceInfoCDC[i].deviceAddress, USB_SETUP_HOST_TO_DEVICE | USB_SETUP_TYPE_CLASS | USB_SETUP_RECIPIENT_INTERFACE, 
                                                            USB_CDC_SET_CONTROL_LINE_STATE , *(uint16_t*)PTR_HOST_CONTROL_LINE_BUFFER , deviceInfoCDC[i].commInterface.interfaceNum , USB_CDC_CONTROL_LINE_LENGTH, NULL,
                                                            USB_DEVICE_REQUEST_SET , deviceInfoCDC[i].clientDriverID ))    
                                       {           
                                            deviceInfoCDC[i].state = STATE_WAIT_FOR_SET_CONTROL_LINE_STATE;
                                       }
                                       else
                                       {
                                            _USBHostCDC_TerminateTransfer(USB_CDC_COMMAND_FAILED);
                                       }
                                    }
                                    else
                                    {
                                        // f[^\zǂłȂꍇ́AsR[fBO̐ݒɐi݂܂B
                                       if(!USBHostIssueDeviceRequest( deviceInfoCDC[i].deviceAddress, USB_SETUP_HOST_TO_DEVICE | USB_SETUP_TYPE_CLASS | USB_SETUP_RECIPIENT_INTERFACE, 
                                                            USB_CDC_SET_LINE_CODING, 1 , deviceInfoCDC[i].commInterface.interfaceNum , USB_CDC_LINE_CODING_LENGTH, (uint8_t*)PTR_HOST_LINE_CODING_BUFFER,
                                                            USB_DEVICE_REQUEST_SET , deviceInfoCDC[i].clientDriverID ))
                                       {           
                                            deviceInfoCDC[i].state = STATE_WAIT_FOR_SET_LINE_CODING;
                                       }
                                       else
                                       {
                                            _USBHostCDC_TerminateTransfer(USB_CDC_COMMAND_FAILED);
                                       }
                                    }
                                }
                            }
                           
                        break;
                    case STATE_WAIT_FOR_SET_LINE_CODING :
                            // sR[fBO̐ݒ
                            if (USBHostTransferIsComplete( deviceInfoCDC[i].deviceAddress, 0, &errorCode, &byteCount ))
                            {
                                if (errorCode)
                                {
                                    /* G[R[hݒ肷 */
                                    _USBHostCDC_TerminateTransfer(errorCode);
                                }
                                else
                                {    
                                    // CR[fBOݒ肳Ă邩ēxmF܂B
                                    // foCXGETCR[fBONGXg𑗂܂ 
                                    if(!USBHostIssueDeviceRequest( deviceInfoCDC[i].deviceAddress, USB_SETUP_DEVICE_TO_HOST | USB_SETUP_TYPE_CLASS | USB_SETUP_RECIPIENT_INTERFACE, 
                                                                 USB_CDC_GET_LINE_CODING, 0 , deviceInfoCDC[i].commInterface.interfaceNum , USB_CDC_LINE_CODING_LENGTH, (uint8_t*)&CDC_DEV_LINE_CODING_Buffer,
                                                                 USB_DEVICE_REQUEST_GET , deviceInfoCDC[i].clientDriverID ))
                                    {           
                                        // ]Cxg҂ĂݒύX邩ǂ肵܂
                                        deviceInfoCDC[i].state     = STATE_WAIT_FOR_GET_LINE_CODING;
                                    }
                                    else
                                    {
                                        _USBHostCDC_TerminateTransfer(USB_CDC_COMMAND_FAILED);
                                    }
                                }
                            }

                        break;

                    case STATE_WAIT_FOR_SET_CONTROL_LINE_STATE :
                        // 䃉C̐ݒ
                        if (USBHostTransferIsComplete( deviceInfoCDC[i].deviceAddress, 0, &errorCode, &byteCount ))
                        {
                            if (errorCode)
                            {
                                /* G[R[hݒ肷 */
                                _USBHostCDC_TerminateTransfer(errorCode);
                            }
                            else
                            {
                                // foCX̓f[^𑗎M鏀ł 
                                USB_HOST_APP_EVENT_HANDLER(deviceInfoCDC[i].deviceAddress,EVENT_CDC_ATTACH,NULL, 0);
                                deviceInfoCDC[i].state     = STATE_RUNNING;
                            }
                        }   

                        break;
                        
                    case STATE_READ_REQ_WAIT :
                              if (USBHostTransferIsComplete( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA, &errorCode, &byteCount ))
                              {
                                if (errorCode)
                                {
                                    if(USB_ENDPOINT_STALLED == errorCode)
                                    {
                                         USBHostClearEndpointErrors( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA );
                                         deviceInfoCDC[i].returnState = STATE_HOLDING;
                                         deviceInfoCDC[i].flags.bfReset = 1;
                                         _USBHostCDC_ResetStateJump( i );
                                    }
                                    else
                                    {
                                        _USBHostCDC_TerminateTransfer(errorCode);
                                    }
                                }
                                else
                                {
                                    // STALLNAĂB EP0ł̂ŁAX̓Xg[NAKv͂܂B
                                    USBHostClearEndpointErrors( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA );
                                    deviceInfoCDC[i].bytesTransferred = byteCount; /* |[gTCYƃtOG[Ɣrł܂ ???*/
                                    deviceInfoCDC[i].state = STATE_RUNNING;
                                    if(deviceInfoCDC[i].endpointDATA == 0x00)
                                    {
                                        USB_HOST_APP_EVENT_HANDLER(deviceInfoCDC[i].deviceAddress,EVENT_CDC_COMM_READ_DONE,NULL, 0);
                                    }
                                    else
                                    {   // f[^C^[tF[Xɑ΂ǂݎv
                                        USB_HOST_APP_EVENT_HANDLER(deviceInfoCDC[i].deviceAddress,EVENT_CDC_DATA_READ_DONE,NULL, 0);
                                    }
                                }
                              }
#ifdef DEBUG_MODE
                              UART2PrintString("|");
#endif                                
                        break;
                    
                    case STATE_WRITE_REQ_WAIT :
                              if (USBHostTransferIsComplete( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA, &errorCode, &byteCount ))
                              {
                                if (errorCode)
                                {
                                    if(USB_ENDPOINT_STALLED == errorCode)
                                    {
                                         USBHostClearEndpointErrors( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA );
                                         deviceInfoCDC[i].returnState = STATE_HOLDING;
                                         deviceInfoCDC[i].flags.bfReset = 1;
                                         _USBHostCDC_ResetStateJump( i );
                                    }
                                    else
                                    {
                                        _USBHostCDC_TerminateTransfer(errorCode);
                                    }
                                }
                                else
                                {
                                    // STALLNAĂB EP0ł̂ŁAX̓Xg[NAKv͂܂B
                                    USBHostClearEndpointErrors( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA );
                                    deviceInfoCDC[i].bytesTransferred = byteCount; /* Can compare with report size and flag error ???*/
                                    deviceInfoCDC[i].state = STATE_RUNNING;
                                    if(deviceInfoCDC[i].endpointDATA == 0x00)
                                    {
                                        USB_HOST_APP_EVENT_HANDLER(deviceInfoCDC[i].deviceAddress,EVENT_CDC_COMM_WRITE_DONE,NULL, 0);
                                    }
                                    else
                                    {   // f[^C^[tF[Xɑ΂ǂݎv
                                        USB_HOST_APP_EVENT_HANDLER(deviceInfoCDC[i].deviceAddress,EVENT_CDC_DATA_WRITE_DONE,NULL, 0);
                                    }
                                }
                              }
#ifdef DEBUG_MODE
                              UART2PrintString("|");
#endif                                
                        break;

                    default:
                            return false;
                }
            #endif
        case EVENT_SOF:         // SOFg[NM܂ - zXgX^bNł͎gp܂
        case EVENT_RESUME:      // foCXE[h̎MĊJ - zXgX^bNł͎gp܂
        case EVENT_SUSPEND:     // foCX[h̃TXyh/AChCxgM - zXgX^bNł͎gp܂
        case EVENT_RESET:       // foCX[h̃oXZbgM - zXgX^bNł͎gp܂
        case EVENT_STALL:       // Xg[܂ - zXgX^bNł͎gp܂
            return true;
            break;
        case EVENT_BUS_ERROR:   // USBɓ]G[܂B
            for (i=0; (i<USB_MAX_CDC_DEVICES) && (deviceInfoCDC[i].deviceAddress != address); i++);
            if (i < USB_MAX_CDC_DEVICES)
            {
                if(transfer_data->bErrorCode == USB_ENDPOINT_NAK_TIMEOUT)
                {
                    USB_HOST_APP_EVENT_HANDLER(deviceInfoCDC[i].deviceAddress,EVENT_CDC_NAK_TIMEOUT,NULL, 0);
                    deviceInfoCDC[i].state = STATE_RUNNING;
                    USBHostClearEndpointErrors( deviceInfoCDC[i].deviceAddress, deviceInfoCDC[i].endpointDATA );
                    return true;
                }
            }
            break;

        default:
            return false;
            break;
    }
    return false;
}

/*******************************************************************************
  Function:
    bool USBHostCDCInitialize( uint8_t address, uint32_t flags, uint8_t clientDriverID )

  Tv:
    ̊֐͂̃NCAghCȍ[`łB

  :
    USBfoCXGj[VĂ鎞ɃzXgwɂČĂяo܂B
    CDCfoCX̏ꍇ́ACDCfBXNv^AC^tF[XfBXNv^A
    yуGh|CgfBXNv^𒲂ׂKv܂B
    ͒ʐMƃf[^C^[tF[X̏fR[h܂B

  Precondition:
    None

  :
    uint8_t address        - VfoCX̃AhX
    uint32_t flags         - tO
    uint8_t clientDriverID - foCXvɑ΂NCAghCo̎ID

  ߂l:
    true   - foCXT|[g邱Ƃł܂B
    false  - T|[gł܂B

  Remarks:
    None
*******************************************************************************/
bool USBHostCDCInitialize( uint8_t address, uint32_t flags, uint8_t clientDriverID )
{
    uint8_t   *descriptor         = NULL;
    uint16_t    i                 = 0;
    uint16_t    endpointINsize    = 0;
    uint16_t    endpointOUTsize   = 0;
    uint8_t    device             = 0;
    uint8_t    endpointIN         = 0;
    uint8_t    endpointOUT        = 0;
    uint8_t    validCommInterface = 0;
    uint8_t    validDataInterface = 0;

    #ifdef DEBUG_MODE
        UART2PrintString( "CDC: USBHostCDCInitialize(0x" );
        UART2PutHex( flags );
        UART2PrintString( ")\r\n" );
    #endif

    // e[uŃfoCX܂AȂꍇ̓G[Ԃ܂B
    #ifdef DEBUG_MODE
        UART2PrintString("CDC: Selecting configuration...\r\n" );
    #endif
    for (device = 0; (device < USB_MAX_CDC_DEVICES) ; device++)
    {
        if(deviceInfoCDC[device].deviceAddress == address)
            return true;
    }

    for (device = 0; (device < USB_MAX_CDC_DEVICES) && (deviceInfoCDC[device].deviceAddress != 0); device++);
    if (device == USB_MAX_CDC_DEVICES)
    {
        #ifdef DEBUG_MODE
            UART2PrintString("CDC: Not in the table!\r\n" );
        #endif
        return false;
    }

    USBHostCDC_Init_CDC_Buffers();
    descriptor = USBHostGetCurrentConfigurationDescriptor( address );

        i = 0;

        #ifdef DEBUG_MODE
            UART2PrintString("CDC: Checking descriptor " );
            UART2PutDec( descriptor[i+5] );
            UART2PrintString(" ...\r\n" );
        #endif

        // C^[tF[X̑
        deviceInfoCDC[device].noOfInterfaces = descriptor[i+4] ;
       
        i += descriptor[i]; // RtBM[VfBXNv^XLbv

        // ̍\Ɍ݂̐ݒݒ肵܂B ŕύXł܂B

        // ̃C^[tF[XfBXNv^܂B
        while (i < ((USB_CONFIGURATION_DESCRIPTOR *)descriptor)->wTotalLength)
        {
            #ifdef DEBUG_MODE
                UART2PrintString("CDC:  Checking interface...\r\n" );
            #endif
            // foCXAhX
            deviceInfoCDC[device].deviceAddress = address;
          
            // C^[tFCXfBXNv^wĂ邩ǂmF܂B
            if (descriptor[i+1] == USB_DESCRIPTOR_INTERFACE)
            {
                // C^[tFCXCDCʐMC^[tFCXǂmF܂B
                if ((descriptor[i+5] == USB_CDC_COMM_INTF) || (descriptor[i+5] == 0xFF))
                {
                    validCommInterface = 1;
                    deviceInfoCDC[device].commInterface.interfaceNum  = descriptor[i+2];
                    deviceInfoCDC[device].commInterface.noOfEndpoints = descriptor[i+4];
                    // ɒB́ATuNXƃvgRzXgɂăT|[gĂ鎖Ӗ܂
                    
                    // NXŗL̋@\fBXNv^T
                     i += descriptor[i]; // C^[tF[XfBXNv^XLbv
                     
                    // SĂ̋@\fBXNv^͂܂
                     while(USB_CDC_CS_INTERFACE == descriptor[i+1] )
                    {
                         switch(descriptor[i+2])
                        {
                             case   USB_CDC_DSC_FN_HEADER   :  
                                         deviceInfoCDC[device].commInterface.Header_Fn_Dsc = 
                                                            (*((USB_CDC_HEADER_FN_DSC*)&descriptor[i]));
                                 break;

                             case   USB_CDC_DSC_FN_CALL_MGT : 
                                         deviceInfoCDC[device].commInterface.Call_Mgt_Fn_Desc = 
                                                            (*((USB_CDC_CALL_MGT_FN_DSC*)&descriptor[i]));
                                 break;

                             case   USB_CDC_DSC_FN_ACM      :
                                         deviceInfoCDC[device].commInterface.ACM_Fn_Desc = 
                                                            (*((USB_CDC_ACM_FN_DSC*)&descriptor[i]));
                                 break;

                             case   USB_CDC_DSC_FN_UNION    :
                                         deviceInfoCDC[device].commInterface.Union_Fn_Desc = 
                                                            (*((USB_CDC_UNION_FN_DSC*)&descriptor[i]));
                                 break;

                             default                        :  // @\̓T|[gĂ܂ - tOG[AfalseԂ܂
                                 break;
                        }
                        
                        i += descriptor[i]; // ̋@\fBXNv^ɐi
                    }

                    if(descriptor[i+1] == USB_DESCRIPTOR_ENDPOINT)
                    {
                        if(descriptor[i+3] == 0x03) // ʐMC^tF[X͊荞݃Gh|Cggp
                        {
                            if (((descriptor[i+2] & 0x80) == 0x80) && (deviceInfoCDC[device].commInterface.endpointIN == 0))
                            {
                                deviceInfoCDC[device].commInterface.endpointIN           = descriptor[i+2];
                                deviceInfoCDC[device].commInterface.endpointMaxDataSize  = ((descriptor[i+4]) |
                                                                                        (descriptor[i+5] << 8));
                                deviceInfoCDC[device].commInterface.endpointPollInterval = descriptor[i+6];
                            }
                        }
                        i += descriptor[i];
                    }
                    else
                    {
                        // ʏAʐMC^tF[Xɂ1̃Gh|Cg܂B 
                        // ʒm̗vf̓IvVł - ݂̓T|[gĂ܂B
                    }

                    if (((descriptor[i+1] == USB_DESCRIPTOR_INTERFACE) && (descriptor[i+5] == USB_CDC_DATA_INTF)) ||
                        (descriptor[i+1] == USB_DESCRIPTOR_ENDPOINT))

                    {
                           if(descriptor[i+5] == USB_CDC_DATA_INTF)
                           {
                              validDataInterface = 1;
                              deviceInfoCDC[device].dataInterface.interfaceNum  = descriptor[i+2];
                              deviceInfoCDC[device].dataInterface.noOfEndpoints = descriptor[i+4];
                              
                              i += descriptor[i]; // Gh|CgfBXNv^֐i
                           }
                           else
                           {
                              // CDCfoCX̒ɂ́ADATAC^[tF[XfBXNv^ʁXɃXgȂ܂B
                              // DATAGh|CgCommunicatonC^[tF[XƈꏏɕʋLĂ܂B
                              validDataInterface = 1;
                              deviceInfoCDC[device].dataInterface.interfaceNum  = deviceInfoCDC[device].commInterface.interfaceNum;
                              deviceInfoCDC[device].dataInterface.noOfEndpoints = deviceInfoCDC[device].commInterface.noOfEndpoints - 1;
                           }
                           // oN܂̓AC\NiXAINOUTGh|CgT܂B
                           endpointIN  = 0;
                           endpointOUT = 0;
                           
                           while(descriptor[i+1] == USB_DESCRIPTOR_ENDPOINT)
                           {   // f[^C^[tF[X̓oN܂͓Gh|Cggp
                               deviceInfoCDC[device].dataInterface.endpointType = descriptor[i+3];
                               if((descriptor[i+3] == 0x01)||(descriptor[i+3] == 0x02))
                               {
                                   if (((descriptor[i+2] & 0x80) == 0x80) && (endpointIN == 0))
                                   {
                                       endpointIN     = descriptor[i+2];
                                       endpointINsize = ((descriptor[i+4])|(descriptor[i+5] << 8));
                                   }
                                   if (((descriptor[i+2] & 0x80) == 0x00) && (endpointOUT == 0))
                                   {
                                       endpointOUT     = descriptor[i+2];
                                       endpointOUTsize = ((descriptor[i+4])|(descriptor[i+5] << 8));
                                   }
                           
                               }
                               i += descriptor[i];
                           }
                           
                           if ((endpointIN != 0) || (endpointOUT != 0)) // ʏAGh|CǵA^CṽyAɂKv܂
                           {
                               deviceInfoCDC[device].dataInterface.endpointIN          = endpointIN;
                               deviceInfoCDC[device].dataInterface.endpointInDataSize  = endpointINsize;
                           
                               deviceInfoCDC[device].dataInterface.endpointOUT         = endpointOUT;
                               deviceInfoCDC[device].dataInterface.endpointOutDataSize = endpointOUTsize;
                           
                               USBHostSetNAKTimeout( address, endpointIN,  USB_HOST_CDC_NAK_TIMEOUT, USB_NUM_BULK_NAKS );
                               USBHostSetNAKTimeout( address, endpointOUT, USB_HOST_CDC_NAK_TIMEOUT, USB_NUM_BULK_NAKS );
                           }
                           else
                           {
                               validDataInterface = 0;
                           }
                    } // if data interface
                } // communication interface loop

            } // if interface descriptor

           // ̐ݒ̎̃fBXNv^ɐi݂܂B
           i += descriptor[i];
        } // main loop - to parse descriptors
        // ACMTuNX͗LȒʐMƃf[^C^tF[XT|[gȂ΂Ȃ܂
        if((validDataInterface == 0) ||(validCommInterface == 0))
        {
            return false;
        }

        deviceInfoCDC[device].clientDriverID = clientDriverID;

        #ifndef USB_ENABLE_TRANSFER_EVENT
           deviceInfoCDC[device].state = STATE_INITIALIZE_DEVICE;
        #else
           // AvP[VpɃfoCXAhX
           USBHostCDCInitAddress( deviceInfoCDC[device].deviceAddress,0,deviceInfoCDC[device].clientDriverID );
           // foCXGET R[fBOv𑗐M
           if(USBHostIssueDeviceRequest( deviceInfoCDC[device].deviceAddress, USB_SETUP_DEVICE_TO_HOST | USB_SETUP_TYPE_CLASS | USB_SETUP_RECIPIENT_INTERFACE, 
                                     USB_CDC_GET_LINE_CODING, 0 , deviceInfoCDC[device].commInterface.interfaceNum , USB_CDC_LINE_CODING_LENGTH, (uint8_t*)&CDC_DEV_LINE_CODING_Buffer,
                                     USB_DEVICE_REQUEST_GET , deviceInfoCDC[device].clientDriverID ))
           {           
                return false;
           }
           // ]Cxg҂ĂݒύX邩ǂ肵܂B
           deviceInfoCDC[device].state = STATE_WAIT_FOR_GET_LINE_CODING;
        #endif
       return true;
}


// *****************************************************************************
// *****************************************************************************
// Internal Functions
// *****************************************************************************
// *****************************************************************************

/*******************************************************************************
  Function:
    void _USBHostCDC_ResetStateJump( uint8_t i )

  Summary:

  :
    ̊֐́AɃZbĝǂ̕sKv邩𔻒fȀԂɃWv܂B

  O:
    foCX́AdeviceInfozɂȂ΂Ȃ܂B

  :
    uint8_t i  -ZbgfoCXdeviceInfoMSD\̂ւ̃CfbNX

  Returns:
    None

  Remarks:
    None
*******************************************************************************/
void _USBHostCDC_ResetStateJump( uint8_t i )
{
     #ifdef USB_ENABLE_TRANSFER_EVENT
        uint8_t    errorCode;
     #endif
   if (deviceInfoCDC[i].flags.bfReset)
    {
        #ifndef USB_ENABLE_TRANSFER_EVENT
            deviceInfoCDC[i].state = STATE_CDC_RESET_RECOVERY;
        #else
             errorCode = USBHostIssueDeviceRequest( deviceInfoCDC[i].deviceAddress, USB_SETUP_HOST_TO_DEVICE | USB_SETUP_TYPE_CLASS | USB_SETUP_RECIPIENT_INTERFACE,
                            USB_CDC_RESET, 0, deviceInfoCDC[i].interface, 0, NULL, USB_DEVICE_REQUEST_SET , deviceInfoCDC[i].clientDriverID );

            if (errorCode)
            {
                // MCHP - Now what??
                _USBHostCDC_TerminateTransfer( USB_CDC_RESET_ERROR );
            }
            else
            {
                deviceInfoCDC[i].state = STATE_WAIT_FOR_RESET;
            }
        #endif
    }
    else
    {
         USB_HOST_APP_EVENT_HANDLER(deviceInfoCDC[i].deviceAddress,EVENT_CDC_RESET,NULL, 0);

         deviceInfoCDC[i].state = deviceInfoCDC[i].returnState;
    }
}



/*******************************************************************************
  Function:
    void USBHostCDC_Init_CDC_Buffers(void)

  Summary:

  :
    ̊֐̓t@Cusb_config.hŃAvP[VɂĐݒ肳ꂽl
    R[fBOobt@܂B
    AvP[V}NCɂȂꍇ́AftHgl蓖Ă܂B
    {[[gAXgbvrbgAf[^rbgApeBrbg̓foCXŊ҂ݒłB
    Gj[VɁAfoCX̒lɐݒ肳ĂȂCDCNCAgoꍇ́A
    CDCNCAg̓foCXKvȒlɐݒ肵܂B

  O:
    ݒ}ŃAusb_config.ht@Cɒ`Kv܂B
    USB Configc[gpCDCACM[hɐݒł܂
  
  Parameters:
    None
  
  Returns:
    None

  l:
    AvP[Vt@Cusb_config.hɃ}N`ĂȂꍇ́A
    t@Cusb_host_cdc.cɒ`ĂftHgll܂B
*******************************************************************************/
void USBHostCDC_Init_CDC_Buffers(void)
{
   CDC_LINE_CODING_Buffer.dwDTERate = USB_CDC_BAUDRATE_SUPPORTED;
   CDC_LINE_CODING_Buffer.bCharFormat = USB_CDC_STOP_BITS;
   CDC_LINE_CODING_Buffer.bParityType = USB_CDC_PARITY_TYPE;
   CDC_LINE_CODING_Buffer.bDataBits   = (uint8_t)USB_CDC_NO_OF_DATA_BITS;
}
