// DOM-IGNORE-BEGIN
/*******************************************************************************
Copyright 2015 Microchip Technology Inc. (www.microchip.com)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

To request to license the code under the MLA license (www.microchip.com/mla_license), 
please contact mla_licensing@microchip.com
*******************************************************************************/
//DOM-IGNORE-END

/*******************************************************************************
 Change History:
 Revision     Description
 v2.7         Modified the code to allow connection of USB-RS232 dongles that do
              not fully comply with CDC specifications
              Modified API USBHostCDC_Api_Send_OUT_Data to allow data transfers
              more than 256 bytes
********************************************************************************/
/*  ̃\[X́Ausb_host_cdc.capp_host_cdc_basicc.cƂ̋nvOłB

    2019-02-05  ޒH[(ނ)  Rg̓{ꉻ(Japanese commentary)
 */

#include <stdlib.h>
#include <string.h>
#include "usb_config.h"
#include "usb.h"
#include "usb_host_cdc.h"
#include "usb_host_cdc_interface.h"


// *****************************************************************************
// *****************************************************************************
// Constants
// *****************************************************************************
// *****************************************************************************

// *****************************************************************************
// State Machine Constants
// *****************************************************************************
//******************************************************************************
//******************************************************************************
// Data Structures
//******************************************************************************
//******************************************************************************

//******************************************************************************
//******************************************************************************
// Section: Local Prototypes
//******************************************************************************
//******************************************************************************
//******************************************************************************
//******************************************************************************
// Macros
//******************************************************************************
//******************************************************************************


//******************************************************************************
//******************************************************************************
// Section: CDC Host Global Variables
//******************************************************************************
//******************************************************************************

//******************************************************************************
//******************************************************************************
// Section: CDC Host External Variables
//******************************************************************************
//******************************************************************************
extern USB_CDC_DEVICE_INFO                  deviceInfoCDC[] __attribute__ ((aligned));
extern uint8_t CDCdeviceAddress ;

//******************************************************************************
//******************************************************************************
// Section: CDC Host Global Variables
//******************************************************************************
//******************************************************************************


/****************************************************************************
  Function:
    bool USBHostCDC_Api_Get_IN_Data(uint8_t no_of_bytes, uint8_t* data)

  :
    ̊֐́ADATAC^tF[Xē̓f[^󂯎ׂɃAvP[VɂČĂяo܂B
@@̋@\́AfoCXf[^Mvݒ肵܂B

  Precondition:
    None

  :
    uint8_t    no_of_bytes - foCX\zoCg(MAX 64)B
    uint8_t*   data        - AvP[VMf[^obt@ւ̃|C^B

  ߂l:
    true    -   ]v͐ɍs܂B
    false   -   ]vs܂B

  Remarks:
    None
***************************************************************************/
bool USBHostCDC_Api_Get_IN_Data(uint8_t no_of_bytes, uint8_t* data)
{
    uint8_t    i;

    for (i=0; (i<USB_MAX_CDC_DEVICES) && (deviceInfoCDC[i].deviceAddress != CDCdeviceAddress); i++);

   if(!USBHostCDCRead_DATA(CDCdeviceAddress, deviceInfoCDC[i].dataInterface.interfaceNum,
                           no_of_bytes,data,deviceInfoCDC[i].dataInterface.endpointIN))
    {
       return true;
    }
    return false;
}


/****************************************************************************
  Function:
    bool USBHostCDC_Api_Send_OUT_Data(uint8_t no_of_bytes, uint8_t* data)

  :
    ̊֐́ADATAC^tF[Xăf[^𑗐MׂɃAvP[VɂČĂяo܂B
    ̋@\́AfoCXւ̃f[^Mvݒ肵܂B

  Precondition:
    None

  Parameters:
    uint8_t    no_of_bytes - foCX\zoCgB
    uint8_t*   data        - AvP[VMf[^obt@ւ̃|C^B


  Return Values:
    true    -   Mv͐ɍs܂B
    false   -   Mvs܂B

  Remarks:
    None
***************************************************************************/
bool USBHostCDC_Api_Send_OUT_Data(uint16_t no_of_bytes, uint8_t* data)
{
    uint8_t    i;

    for (i=0; (i<USB_MAX_CDC_DEVICES) && (deviceInfoCDC[i].deviceAddress != CDCdeviceAddress); i++);

   if(!USBHostCDCSend_DATA(CDCdeviceAddress, deviceInfoCDC[i].dataInterface.interfaceNum,
                           no_of_bytes,data,deviceInfoCDC[i].dataInterface.endpointOUT))
    {
       return true;
    }
    return false;
}


/****************************************************************************
  Function:
    bool USBHostCDC_ApiTransferIsComplete(uint8_t* errorCodeDriver,uint8_t* byteCount)

  :
    ̊֐́A]Xe[^X𒲂ׂׂɃAvP[VɂČĂяo܂B
    ]IƁÅ֐trueԂ܂B
    ]ǂmFɂ́AAvP[V͎QƂɂĕԂꂽG[R[h
    mFKv܂B

  Precondition:
    None

  :
    uint8_t    *errorCodeDriver - ߂l(G[R[h)
    uint8_t    *byteCount       - ]ꂽoCg


  ߂l:
    true    -   ]͊܂B
    false   -   ]͕ۗłB

  Remarks:
    None
***************************************************************************/
bool USBHostCDC_ApiTransferIsComplete(uint8_t* errorCodeDriver, uint8_t* byteCount )
{       
    return(USBHostCDCTransferIsComplete(CDCdeviceAddress,errorCodeDriver,byteCount));
}

/*******************************************************************************
  Function:
    bool USBHostCDC_ApiDeviceDetect( void )

  :
    ̋@\́ACDCfoCXڑĂĎgp\ǂ𔻒f܂B

  Precondition:
    None

  Parameters:
    None

  Return Values:
    true   -  CDC݂AłĂ
    false  -  CDC݂Ȃ͏łĂ܂

  l:
    ̓foCX҂ԂɃ[vŌĂяo鎖̂ŁA^XNsĂ鎖mF܂B
*******************************************************************************/
bool USBHostCDC_ApiDeviceDetect( void )
{
    USBHostTasks();
    USBHostCDCTasks();

    if ((USBHostCDCDeviceStatus(CDCdeviceAddress) == USB_CDC_NORMAL_RUNNING) &&
        (CDCdeviceAddress != 0))
    {
        return true;
    }

    return false;
}


/*******************************************************************************
  Function:
    uint8_t USBHostCDC_Api_ACM_Request(uint8_t requestType, uint8_t size, uint8_t* data)

  :
    ̋@\AvP[VR[hŎgpāAACM(Abstract Control Model)ŗL̗vɓIɃANZX鎖ł܂B
    ̋@\́AAvP[VOɐݒ肳ꂽ[g{[[gύX悤ƂĂꍇɂ̂ݎgpĂB
    foCXƂ̊ԂőMf[^̓oCg̔złB
    AvP[V͂̋@\gpOɃf[^tH[}bg𗝉鎖ɓʂȒӂ𕥂Kv܂B

  O:
    foCXGj[VĐڑKv܂B

  :
    uint8_t    requestType  - USB_CDC_SEND_ENCAPSULATED_COMMAND
                            - USB_CDC_GET_ENCAPSULATED_REQUEST
                            - USB_CDC_SET_LINE_CODING
                            - USB_CDC_SET_CONTROL_LINE_STATE
                              ́ACDCzXgX^bNɂăT|[gĂK{CDCvłB
    uint8_t    size         - ]oCg
    uint8_t    *data        - ]Ăf[^ւ̃|C^B

  ߂l:
    USB_SUCCESS                 - NGXg͐ɊJn܂
    USB_CDC_DEVICE_NOT_FOUND    - w肳ꂽAhXfoCX܂
    USB_CDC_DEVICE_BUSY         - ]sׂ̃foCXK؂ȏԂł͂܂
    USB_CDC_COMMAND_FAILED      - v̓T|[gĂ܂B
    USB_CDC_ILLEGAL_REQUEST     - vꂽIDłB

  Remarks:
    None
 *******************************************************************************/
uint8_t USBHostCDC_Api_ACM_Request(uint8_t requestType, uint8_t size, uint8_t* data)
{
    uint8_t    i;
    uint8_t    return_val = USB_CDC_COMMAND_FAILED;

    // Find the correct device.
    for (i=0; (i<USB_MAX_CDC_DEVICES) && (deviceInfoCDC[i].deviceAddress != CDCdeviceAddress); i++);
    if (i == USB_MAX_CDC_DEVICES)
    {
        return USB_CDC_DEVICE_NOT_FOUND;
    }

    switch(requestType)
    {
        case USB_CDC_SEND_ENCAPSULATED_COMMAND : 
                               return_val = USBHostCDCTransfer(CDCdeviceAddress, USB_CDC_SEND_ENCAPSULATED_COMMAND,0,
                                                       deviceInfoCDC[i].commInterface.interfaceNum,size,data,
                                                       0);
            break;
        case USB_CDC_GET_ENCAPSULATED_REQUEST :
                               return_val = USBHostCDCTransfer(CDCdeviceAddress,  USB_CDC_GET_ENCAPSULATED_REQUEST,1,
                                                       deviceInfoCDC[i].commInterface.interfaceNum,size,data,
                                                       0);
            break;

        case USB_CDC_SET_COMM_FEATURE :
            break;

        case USB_CDC_GET_COMM_FEATURE :
            break;
        
        case USB_CDC_SET_LINE_CODING :
                               return_val = USBHostCDCTransfer(CDCdeviceAddress,  USB_CDC_SET_LINE_CODING,0,
                                                       deviceInfoCDC[i].commInterface.interfaceNum,size,data,
                                                       0);
            break;
        
        case USB_CDC_GET_LINE_CODING :
                               return_val = USBHostCDCTransfer(CDCdeviceAddress,  USB_CDC_GET_LINE_CODING,1,
                                                       deviceInfoCDC[i].commInterface.interfaceNum,size,data,
                                                       0);
            break;
        
        case USB_CDC_SET_CONTROL_LINE_STATE :
                               return_val = USBHostCDCTransfer(CDCdeviceAddress,  USB_CDC_SET_CONTROL_LINE_STATE,0,
                                                       deviceInfoCDC[i].commInterface.interfaceNum,size,data,
                                                       0);
            break;

        case USB_CDC_SEND_BREAK :
            break;
        default:
                 return USB_CDC_ILLEGAL_REQUEST;
            break;
    }
    return return_val;
}
